package main

import "core:fmt"
import "core:strings"

LLVMScope :: struct {
    name: cstring,
    basic_block: LLVMBasicBlockRef,
    definitions: map[int]LLVMValueRef,
    types: map[int]LLVMTypeRef,
}

g_llvm_scope_stack := [dynamic]LLVMScope{}

llvm_scope_find_definition :: proc(name: ^[dynamic]u8) -> LLVMValueRef {
    #reverse for &scope in g_llvm_scope_stack {
        value, ok := scope.definitions[get_character_sum_of_dyn_arr(name)]
        if ok {
            return value
        }
    }
    return nil
}

llvm_scope_find_type :: proc(name: ^[dynamic]u8) -> LLVMTypeRef {
    #reverse for &scope in g_llvm_scope_stack {
        value, ok := scope.types[get_character_sum_of_dyn_arr(name)]
        if ok {
            return value
        }
    }
    return nil
}

llvm_scope_enter :: proc(name: cstring, basic_block: LLVMBasicBlockRef) {
    append(&g_llvm_scope_stack, LLVMScope{
        name = name,
        basic_block = basic_block,
        definitions = make(map[int]LLVMValueRef),
        types = make(map[int]LLVMTypeRef),
    })
}

llvm_scope_leave :: proc() {
    delete(g_llvm_scope_stack[len(g_llvm_scope_stack)-1].definitions)
    delete(g_llvm_scope_stack[len(g_llvm_scope_stack)-1].types)
    pop(&g_llvm_scope_stack)
}

llvm_top_scope :: proc() -> ^LLVMScope {
    return &g_llvm_scope_stack[len(g_llvm_scope_stack)-1]
}

generate_llvm :: proc(ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef, node: ^Node) {
    // FIXME: Everything leaks memory.
    scope_number : i32 = 0

	int_32_type := LLVMInt32TypeInContext(ctx)
	main_function_type := LLVMFunctionType(int_32_type, nil, 0, LLVMBool(0))
	main_function := LLVMAddFunction(mod, "main", main_function_type)

    get_basic_block_name :: proc(scope_num: i32) -> cstring {
        return fmt.caprintf("scope_%d", scope_num)
    }

    generate_llvm_integer :: proc(ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef, node: ^Node) -> LLVMValueRef {
        bit_size := uint(node.return_type.bit_size)
        value := node.value.(u64)
        is_signed := node.return_type.is_signed
        return LLVMConstInt(LLVMIntTypeInContext(ctx, bit_size), value, LLVMBool(is_signed))
    }

    generate_llvm_float :: proc(ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef, node: ^Node) -> LLVMValueRef {
        bit_size := uint(node.return_type.bit_size)
        value := node.value.(f64)
        if bit_size == 64 {
            return LLVMConstReal(LLVMDoubleTypeInContext(ctx), value)
        } else if bit_size == 32 {
            return LLVMConstReal(LLVMFloatTypeInContext(ctx), value)
        }

        append(
            &g_message_list,
            message_create(
                .Error,
                fmt.aprintf("Unsupported float bit size: %d", bit_size),
                node.range,
            ),
        )
        return nil
    }

    generate_llvm_type_from_node :: proc(ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef, type: ^Type) -> LLVMTypeRef {
        if type == nil {
            return LLVMVoidTypeInContext(ctx)
        }

        switch type.kind {
        case .Integer:
            return LLVMIntTypeInContext(ctx, uint(type.bit_size))
        case .Float:
            if type.bit_size == 64 {
                return LLVMDoubleTypeInContext(ctx)
            } else if type.bit_size == 32 {
                return LLVMFloatTypeInContext(ctx)
            } else {
                fmt.panicf("LLVM-IR: Unsupported float type bit size: {}", type.bit_size)
            }
        case .Array:
            array_of := generate_llvm_type_from_node(ctx, mod, builder, type.array_of)
            return LLVMArrayType2(array_of, type.array_size)
        case .Pointer:
            pointer_of := generate_llvm_type_from_node(ctx, mod, builder, type.pointer_to)
            return LLVMPointerType(pointer_of, 0)
        }
        panic("LLVM-IR: Invalid type")
    }

    generate_llvm_function_extern :: proc(
        ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef,
        fn: ^Node, scope_number: ^i32
    ) -> (type: LLVMTypeRef, value: LLVMValueRef) {
        function_args_type := [dynamic]LLVMTypeRef{}
        for arg in fn.children {
            if arg.kind != .VariableDeclaration {
                continue
            }
            append(&function_args_type, generate_llvm_type_from_node(ctx, mod, builder, arg.return_type))
        }
        function_return_type := generate_llvm_type_from_node(ctx, mod, builder, fn.return_type)
        
        type = LLVMFunctionType(function_return_type, raw_data(function_args_type[:]), len(function_args_type), LLVMBool(0))
        value = LLVMAddFunction(mod, strings.clone_to_cstring(string(fn.value.([dynamic]u8)[:])), type)
        return
    }

    generate_llvm_function_call :: proc(
        ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef,
        node: ^Node,
    ) -> LLVMValueRef {
        // FIXME: Add support for dot access
        name := &node.children[0].value.([dynamic]u8)
        fn_type := llvm_scope_find_type(name)
        fn_value := llvm_scope_find_definition(name)
        if fn_type == nil || fn_value == nil {
            append(
                &g_message_list,
                message_create(
                    .Error,
                    fmt.aprintf("Function '%s' not found", name),
                    node.range,
                ),
            )
            return nil
        }

        fn_args := [dynamic]LLVMValueRef{}
        for &arg in node.children[1:] {
            value := generate_llvm_expression(ctx, mod, builder, arg)
            append(&fn_args, value)
        }

        call := LLVMBuildCall2(builder, fn_type, fn_value, raw_data(fn_args[:]), len(fn_args), cstring(raw_data(name[:])))
        return call
    }

    generate_llvm_value :: proc(
        ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef,
        node: ^Node,
    ) -> LLVMValueRef {
        if node.kind == .Integer {
            return generate_llvm_integer(ctx, mod, builder, node)
        } else if node.kind == .Float {
            return generate_llvm_float(ctx, mod, builder, node)
        } else if node.kind == .FunctionCall {
            return generate_llvm_function_call(ctx, mod, builder, node)
        } else if node.kind == .Identifier {
            def := llvm_scope_find_definition(&node.value.([dynamic]u8))
            type := llvm_scope_find_type(&node.value.([dynamic]u8))
            //LLVMBuildLoad2 :: proc(Builder: LLVMBuilderRef, Ty: LLVMTypeRef, PointerVal: LLVMValueRef, Name: cstring) -> LLVMValueRef ---
            def_value := LLVMBuildLoad2(builder, type, def, "loadtmp")
            return def_value
        }

        fmt.panicf("FIXME: Implement other node kinds. Got: {}", node.kind)
    }

    generate_llvm_expression :: proc(
        ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef,
        node: ^Node,
    ) -> LLVMValueRef {
        if node.kind == .BinaryExpression {
            return generate_llvm_binary_expression(ctx, mod, builder, node)
        }
        if node.kind == .Integer || node.kind == .Float || node.kind == .FunctionCall || node.kind == .Identifier {
            return generate_llvm_value(ctx, mod, builder, node)
        }
        if node.kind == .FunctionCall {
            return generate_llvm_function_call(ctx, mod, builder, node)
        }
        return nil
    }

    generate_llvm_binary_expression :: proc(
        ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef,
        node: ^Node,
    ) -> LLVMValueRef {
        op := node.value_token_kind
        lhs_node := node.children[0]
        rhs_node := node.children[1]
        lhs := generate_llvm_expression(ctx, mod, builder, lhs_node)
        rhs := generate_llvm_expression(ctx, mod, builder, rhs_node)
        if lhs_node.return_type.kind != rhs_node.return_type.kind {
            fmt.panicf("LLVM-C: Binary expression types do not match: {} and {}", lhs_node.return_type.kind, rhs_node.return_type.kind)
        }
        is_float := lhs_node.return_type.kind == .Float
        is_signed := lhs_node.return_type.is_signed
        #partial switch op {
        case .Add:
            if is_float {
                return LLVMBuildFAdd(builder, lhs, rhs, "addtmp")
            } else {
                return LLVMBuildAdd(builder, lhs, rhs, "addtmp")
            }
        case .Subtract:
            if is_float {
                return LLVMBuildFSub(builder, lhs, rhs, "subtmp")
            } else {
                return LLVMBuildSub(builder, lhs, rhs, "subtmp")
            }
        case .Multiply:
            if is_float {
                return LLVMBuildFMul(builder, lhs, rhs, "multmp")
            } else {
                return LLVMBuildMul(builder, lhs, rhs, "multmp")
            }
        case .Divide:
            if is_float {
                return LLVMBuildFDiv(builder, lhs, rhs, "divtmp")
            } else {
                if is_signed {
                    return LLVMBuildSDiv(builder, lhs, rhs, "divtmp")
                } else {
                    return LLVMBuildUDiv(builder, lhs, rhs, "divtmp")
                }
            }
        case .Modulo:
            if is_float {
                return LLVMBuildFRem(builder, lhs, rhs, "modtmp")
            } else {
                if is_signed {
                    return LLVMBuildSRem(builder, lhs, rhs, "modtmp")
                } else {
                    return LLVMBuildURem(builder, lhs, rhs, "modtmp")
                }
            }
        case:
            fmt.panicf("LLVM-C: Unexpected binary expression kind: {}", op)
        }
    }

    generate_llvm_scope :: proc(ctx: LLVMContextRef, mod: LLVMModuleRef, builder: LLVMBuilderRef, function: LLVMValueRef, scope: ^Node, scope_number: ^i32) {
        scope_name := get_basic_block_name(scope_number^)
        basic_block := LLVMAppendBasicBlockInContext(ctx, function, scope_name)
        llvm_scope_enter(scope_name, basic_block)
        LLVMPositionBuilderAtEnd(builder, basic_block)
        scope_number^ += 1
        for node in scope.children {
            #partial switch node.kind {
            case .Function:
                panic("FIXME: Implement function delarations")
            case .ExternFunction:
                type, value := generate_llvm_function_extern(ctx, mod, builder, node, scope_number)
                //name := LLVMGetValueName2(value, nil)
                llvm_top_scope().definitions[get_character_sum_of_dyn_arr(&node.value.([dynamic]u8))] = value
                llvm_top_scope().types[get_character_sum_of_dyn_arr(&node.value.([dynamic]u8))] = type
            case .Integer: fallthrough
            case .Float: fallthrough
            case .Identifier: fallthrough
            case .FunctionCall: fallthrough
            case .BinaryExpression:
                generate_llvm_expression(ctx, mod, builder, node)
            case .VariableDeclaration:
                type := generate_llvm_type_from_node(ctx, mod, builder, node.return_type)
                var := LLVMBuildAlloca(builder, type, strings.clone_to_cstring(string(node.children[0].value.([dynamic]u8)[:])))
                if node.children[2] != nil {
                    value := generate_llvm_expression(ctx, mod, builder, node.children[2])
                    LLVMBuildStore(builder, value, var)
                }
                llvm_top_scope().definitions[get_character_sum_of_dyn_arr(&node.children[0].value.([dynamic]u8))] = var
                llvm_top_scope().types[get_character_sum_of_dyn_arr(&node.children[0].value.([dynamic]u8))] = type
            case:
                panic("FIXME: Implement other node kinds")
            }
        }
        llvm_scope_leave()
    }

    generate_llvm_scope(ctx, mod, builder, main_function, node, &scope_number)
    //basic_block := LLVMAppendBasicBlockInContext(ctx, main_function, "end")
    //LLVMPositionBuilderAtEnd(builder, basic_block)
    LLVMBuildRet(builder, LLVMConstInt(int_32_type, 0, LLVMBool(1)))
}
